@@ -4,87 +4,99 @@
     DrawingContexts are able to be handled directly by the UI system or
     produce javascript or svg to do the drawing.
 """
+from __future__ import annotations
 
 # standard libraries
 import base64
 import collections
-from contextlib import contextmanager
+import contextlib
 import copy
+import enum
 import io
 import logging
 import math
 import re
 import struct
 import sys
-import time
 import threading
+import time
 import typing
 import xml.sax.saxutils
 
 # third party libraries
-import imageio
+import imageio.v3 as imageio
 import numpy
 
 # local libraries
-# None
+from nion.utils import Color
+from nion.utils import Geometry
+
 
 # pylint: disable=star-args
 
 
-def get_rgba_view_from_rgba_data(rgba_data):
+RGBA32Type = typing.Any
+RGBA8Type = typing.Any
+GrayscaleF32Type = typing.Any
+C8Type = typing.Any
+
+ByteOrderType = typing.Optional[typing.Union[typing.Literal["little"], typing.Literal["big"]]]
+
+
+def get_rgba_view_from_rgba_data(rgba_data: RGBA32Type) -> RGBA8Type:
     return rgba_data.view(numpy.uint8).reshape(rgba_data.shape + (4,))
 
 
-def get_rgba_data_from_rgba(rgba_image):
+def get_rgba_data_from_rgba(rgba_image: RGBA8Type) -> RGBA32Type:
     return rgba_image.view(numpy.uint32).reshape(rgba_image.shape[:-1])
 
 
-def get_byte_view(rgba_image):
+def get_byte_view(rgba_image: RGBA32Type) -> RGBA8Type:
     return rgba_image.view(numpy.uint8).reshape(rgba_image.shape + (-1, ))
 
 
-def get_red_view(rgba_image, byteorder=None):
+def get_red_view(rgba_image: RGBA32Type, byteorder: ByteOrderType = None) -> C8Type:
     if byteorder is None:
-        byteorder = sys.byteorder
+        byteorder = typing.cast(ByteOrderType, sys.byteorder)
     bytes = get_byte_view(rgba_image)
     assert bytes.shape[-1] == 4
     if byteorder == 'little':
-        return bytes[..., 2]  # strip A off BGRA
+        return typing.cast(C8Type, bytes[..., 2])  # strip A off BGRA
     else:
-        return bytes[..., 1]  # strip A off ARGB
+        return typing.cast(C8Type, bytes[..., 1])  # strip A off ARGB
 
 
-def get_green_view(rgba_image, byteorder=None):
+def get_green_view(rgba_image: RGBA32Type, byteorder: ByteOrderType = None) -> C8Type:
     if byteorder is None:
-        byteorder = sys.byteorder
+        byteorder = typing.cast(ByteOrderType, sys.byteorder)
     bytes = get_byte_view(rgba_image)
     assert bytes.shape[-1] == 4
     if byteorder == 'little':
-        return bytes[..., 1]  # strip A off BGRA
+        return typing.cast(C8Type, bytes[..., 1])  # strip A off BGRA
     else:
-        return bytes[..., 2]  # strip A off ARGB
+        return typing.cast(C8Type, bytes[..., 2])  # strip A off ARGB
 
 
-def get_blue_view(rgba_image, byteorder=None):
+def get_blue_view(rgba_image: RGBA32Type, byteorder: ByteOrderType = None) -> C8Type:
     if byteorder is None:
-        byteorder = sys.byteorder
+        byteorder = typing.cast(ByteOrderType, sys.byteorder)
     bytes = get_byte_view(rgba_image)
     assert bytes.shape[-1] == 4
     if byteorder == 'little':
-        return bytes[..., 0]  # strip A off BGRA
+        return typing.cast(C8Type, bytes[..., 0])  # strip A off BGRA
     else:
-        return bytes[..., 3]  # strip A off ARGB
+        return typing.cast(C8Type, bytes[..., 3])  # strip A off ARGB
 
 
-def get_alpha_view(rgba_image, byteorder=None):
+def get_alpha_view(rgba_image: RGBA32Type, byteorder: ByteOrderType = None) -> C8Type:
     if byteorder is None:
-        byteorder = sys.byteorder
+        byteorder = typing.cast(ByteOrderType, sys.byteorder)
     bytes = get_byte_view(rgba_image)
     assert bytes.shape[-1] == 4
     if byteorder == 'little':
-        return bytes[..., 3]  # A of BGRA
+        return typing.cast(C8Type, bytes[..., 3])  # A of BGRA
     else:
-        return bytes[..., 0]  # A of ARGB
+        return typing.cast(C8Type, bytes[..., 0])  # A of ARGB
 
 
 class DrawingContext:
@@ -99,31 +111,31 @@ class DrawingContext:
     __image_id = 0
     __image_id_lock = threading.RLock()
 
-    def __init__(self):
-        self.commands = []
+    def __init__(self) -> None:
+        self.commands: typing.List[typing.Sequence[typing.Any]] = []
         self.binary_commands = bytearray()
         self.save_count = 0
-        self.images = dict()
+        self.images: typing.Dict[str, RGBA32Type] = dict()
 
-    def copy_from(self, drawing_context):
+    def copy_from(self, drawing_context: DrawingContext) -> None:
         assert self.save_count == 0
         assert drawing_context.save_count == 0
         self.commands = drawing_context.commands
         self.binary_commands = drawing_context.binary_commands
         self.images = drawing_context.images
 
-    def add(self, drawing_context):
+    def add(self, drawing_context: DrawingContext) -> None:
         self.commands.extend(drawing_context.commands)
         self.binary_commands.extend(drawing_context.binary_commands)
         self.images.update(drawing_context.images)
 
-    def clear(self):
+    def clear(self) -> None:
         self.commands = []
-        self.binary_commands = []
+        self.binary_commands = bytearray()
         self.save_count = 0
         self.images = dict()
 
-    def to_js(self):
+    def to_js(self) -> str:
         js = ""
         for command in self.commands:
             command_id = command[0]
@@ -210,31 +222,32 @@ class DrawingContext:
                 js += "{0}.addColorStop({1}, '{2}');".format(js_var, x, color)
         return js
 
-    def to_svg(self, size, viewbox):
+    def to_svg(self, size: Geometry.IntSize, viewbox: Geometry.IntRect) -> str:
         svg = ""
         defs = ""
         path = ""
         next_clip_id = 1
-        transform = list()
-        closers = list()
-        fill_style = None
+        transform: typing.List[str] = list()
+        closers: typing.List[str] = list()
+        fill_style: typing.Optional[str] = None
         fill_opacity = 1.0
-        stroke_style = None
+        stroke_style: typing.Optional[str] = None
         stroke_opacity = 1.0
         line_cap = "square"
         line_join = "bevel"
         line_width = 1.0
-        line_dash = None
+        line_dash: typing.Optional[int] = None
         text_anchor = "start"
         text_baseline = "alphabetic"
-        font_style = None
-        font_weight = None
-        font_size = None
-        font_unit = None
-        font_family = None
-        contexts = collections.deque()
-        gradient_start = None
-        gradient_stops = list()
+        font_style: typing.Optional[str] = None
+        font_weight: typing.Optional[str] = None
+        font_size: typing.Optional[int] = None
+        font_unit: typing.Optional[str] = None
+        font_family: typing.Optional[str] = None
+        # Python 3.9+: collections.deque[typing.Dict[str, typing.Any]]
+        contexts: typing.Any = collections.deque()
+        gradient_start: typing.Optional[str] = None
+        gradient_stops: typing.List[str] = list()
 
         # make a SVG 1.1 compatible color, opacity tuple
         def parse_color(color_str: str) -> typing.Tuple[str, float]:
@@ -249,7 +262,7 @@ class DrawingContext:
             #logging.debug(command_id)
             command_args = command[1:]
             if command_id == "save":
-                context = dict()
+                context: typing.Dict[str, typing.Any] = dict()
                 context["path"] = path
                 context["transform"] = copy.deepcopy(transform)
                 context["fill_style"] = fill_style
@@ -335,7 +348,10 @@ class DrawingContext:
                 w, h, image, image_id, a, b, c, d = command_args
                 png_file = io.BytesIO()
                 rgba_data = get_rgba_view_from_rgba_data(image)
-                imageio.imwrite(png_file, rgba_data[..., (2,1,0,3)], "png")
+                # image compression is time consuming. pass parameters to make this step as fast as possible.
+                # see nionswift-642.
+                # TODO: fix typing when imageio gets their numpy typing correct.
+                imageio.imwrite(png_file, rgba_data[..., (2, 1, 0, 3)], extension=".png", optimize=False, compress_level=1)  # type: ignore
                 png_encoded = base64.b64encode(png_file.getvalue()).decode('utf=8')
                 transform_str = " transform='{0}'".format(" ".join(transform)) if len(transform) > 0 else ""
                 svg_format_str = "<image x='{0}' y='{1}' width='{2}' height='{3}' xlink:href='data:image/png;base64,{4}'{5} />"
@@ -345,14 +361,14 @@ class DrawingContext:
                 m = 255.0 / (high - low) if high != low else 1
                 image = numpy.empty(data.shape, numpy.uint32)
                 if color_table is not None:
-                    adj_color_table = numpy.empty(color_table.shape, numpy.uint32)
+                    adj_color_table: numpy.typing.NDArray[numpy.uint32] = numpy.empty(color_table.shape, numpy.uint32)
                     # ordering of color_table is BGRA
                     # ordering of adj_color_table is RGBA
                     get_byte_view(adj_color_table)[:, 0] = get_byte_view(color_table)[:, 2]
                     get_byte_view(adj_color_table)[:, 1] = get_byte_view(color_table)[:, 1]
                     get_byte_view(adj_color_table)[:, 2] = get_byte_view(color_table)[:, 0]
                     get_byte_view(adj_color_table)[:, 3] = get_byte_view(color_table)[:, 3]
-                    clipped_array = numpy.clip((m * (data - low)).astype(numpy.int), 0, 255).astype(numpy.uint8)
+                    clipped_array = numpy.clip((m * (data - low)).astype(int), 0, 255).astype(numpy.uint8)
                     image[:] = adj_color_table[clipped_array]
                 else:
                     clipped_array = numpy.clip(data, low, high)
@@ -363,7 +379,10 @@ class DrawingContext:
                     get_blue_view(image)[:] = clipped_array
                     get_alpha_view(image)[:] = 255
                 png_file = io.BytesIO()
-                imageio.imwrite(png_file, get_rgba_view_from_rgba_data(image), "png")
+                # image compression is time consuming. pass parameters to make this step as fast as possible.
+                # see nionswift-642.
+                # TODO: fix typing when imageio gets their numpy typing correct.
+                imageio.imwrite(png_file, get_rgba_view_from_rgba_data(image), extension=".png", optimize=False, compress_level=1)  # type: ignore
                 png_encoded = base64.b64encode(png_file.getvalue()).decode('utf=8')
                 transform_str = " transform='{0}'".format(" ".join(transform)) if len(transform) > 0 else ""
                 svg_format_str = "<image x='{0}' y='{1}' width='{2}' height='{3}' xlink:href='data:image/png;base64,{4}'{5} />"
@@ -400,6 +419,7 @@ class DrawingContext:
                                              transform_str)
             elif command_id == "fillStyleGradient":
                 command_var = command_args[0]
+                assert gradient_start is not None
                 defs += gradient_start + "".join(gradient_stops) + "</linearGradient>"
                 fill_style = "url(#{0})".format("grad" + str(command_var))
             elif command_id == "fillStyle":
@@ -469,69 +489,78 @@ class DrawingContext:
         result += "</svg>"
         return result
 
-    @contextmanager
-    def saver(self):
+    @contextlib.contextmanager
+    def saver(self) -> typing.Iterator[typing.Any]:
         self.save()
         try:
             yield
         finally:
            self.restore()
 
-    def save(self):
+    def save(self) -> None:
         self.commands.append(("save", ))
         self.binary_commands.extend(b"save")
         self.save_count += 1
 
-    def restore(self):
+    def restore(self) -> None:
         self.commands.append(("restore", ))
         self.binary_commands.extend(b"rest")
         self.save_count -= 1
 
-    def begin_layer(self, layer_id: int, layer_seed: int, a, b, c, d) -> None:
+    def begin_layer(self, layer_id: int, layer_seed: int, a: float, b: float, c: float, d: float) -> None:
         self.commands.append(("begin_layer", int(layer_id), int(layer_seed), float(a), float(b), float(c), float(d)))
         self.binary_commands.extend(struct.pack("4siiffff", b"bgly", int(layer_id), int(layer_seed), float(a), float(b), float(c), float(d)))
 
-    def end_layer(self, layer_id: int, layer_seed: int, a, b, c, d) -> None:
+    def end_layer(self, layer_id: int, layer_seed: int, a: float, b: float, c: float, d: float) -> None:
         self.commands.append(("end_layer", int(layer_id), int(layer_seed), float(a), float(b), float(c), float(d)))
         self.binary_commands.extend(struct.pack("4siiffff", b"enly", int(layer_id), int(layer_seed), float(a), float(b), float(c), float(d)))
 
-    def begin_path(self):
+    def begin_path(self) -> None:
         self.commands.append(("beginPath", ))
         self.binary_commands.extend(b"bpth")
 
-    def close_path(self):
+    def close_path(self) -> None:
         self.commands.append(("closePath", ))
         self.binary_commands.extend(b"cpth")
 
-    def clip_rect(self, a, b, c, d):
+    def add_path(self, path: Path) -> None:
+        path.add_commands(self)
+
+    def clip_rect(self, a: float, b: float, c: float, d: float) -> None:
         self.commands.append(("clip", float(a), float(b), float(c), float(d)))
         self.binary_commands.extend(struct.pack("4sffff", b"clip", float(a), float(b), float(c), float(d)))
 
-    def translate(self, x, y):
+    def translate(self, x: float, y: float) -> None:
         self.commands.append(("translate", float(x), float(y)))
         self.binary_commands.extend(struct.pack("4sff", b"tran", float(x), float(y)))
 
-    def scale(self, x, y):
+    def scale(self, x: float, y: float) -> None:
         self.commands.append(("scale", float(x), float(y)))
         self.binary_commands.extend(struct.pack("4sff", b"scal", float(x), float(y)))
 
-    def rotate(self, radians):
+    def rotate(self, radians: float) -> None:
         self.commands.append(("rotate", math.degrees(float(radians))))
         self.binary_commands.extend(struct.pack("4sf", b"rota", math.degrees(float(radians))))
 
-    def move_to(self, x, y):
+    def move_to(self, x: float, y: float) -> None:
         self.commands.append(("moveTo", float(x), float(y)))
         self.binary_commands.extend(struct.pack("4sff", b"move", float(x), float(y)))
 
-    def line_to(self, x, y):
+    def line_to(self, x: float, y: float) -> None:
         self.commands.append(("lineTo", float(x), float(y)))
         self.binary_commands.extend(struct.pack("4sff", b"line", float(x), float(y)))
 
-    def rect(self, l, t, w, h):
+    # optimization; use with care.
+    def _line_to_multi(self, line_commands: typing.Sequence[typing.Tuple[float, float]]) -> None:
+        for x, y in line_commands:
+            self.commands.append(("lineTo", x, y))
+            self.binary_commands.extend(struct.pack("4sff", b"line", x, y))
+
+    def rect(self, l: float, t: float, w: float, h: float) -> None:
         self.commands.append(("rect", float(l), float(t), float(w), float(h)))
         self.binary_commands.extend(struct.pack("4sffff", b"rect", float(l), float(t), float(w), float(h)))
 
-    def round_rect(self, x, y, w, h, r):
+    def round_rect(self, x: float, y: float, w: float, h: float, r: float) -> None:
         self.move_to(x + r, y)
         self.arc_to(x + w, y, x + w, y + r, r)
         self.arc_to(x + w, y + h, x + w - r, y + h, r)
@@ -539,23 +568,23 @@ class DrawingContext:
         self.arc_to(x, y, x + r, y, r)
         self.close_path()
 
-    def arc(self, x, y, r, sa, ea, ac=False):
+    def arc(self, x: float, y: float, r: float, sa: float, ea: float, ac: bool = False) -> None:
         self.commands.append(("arc", float(x), float(y), float(r), float(sa), float(ea), bool(ac)))
         self.binary_commands.extend(struct.pack("4sfffffi", b"arc ", float(x), float(y), float(r), float(sa), float(ea), bool(ac)))
 
-    def arc_to(self, x1, y1, x2, y2, r):
+    def arc_to(self, x1: float, y1: float, x2: float, y2: float, r: float) -> None:
         self.commands.append(("arcTo", float(x1), float(y1), float(x2), float(y2), float(r)))
         self.binary_commands.extend(struct.pack("4sfffff", b"arct", float(x1), float(y1), float(x2), float(y2), float(r)))
 
-    def bezier_curve_to(self, x1, y1, x2, y2, x, y):
+    def bezier_curve_to(self, x1: float, y1: float, x2: float, y2: float, x: float, y: float) -> None:
         self.commands.append(("cubicTo", float(x1), float(y1), float(x2), float(y2), float(x), float(y)))
         self.binary_commands.extend(struct.pack("4sffffff", b"cubc", float(x1), float(y1), float(x2), float(y2), float(x), float(y)))
 
-    def quadratic_curve_to(self, x1, y1, x, y):
+    def quadratic_curve_to(self, x1: float, y1: float, x: float, y: float) -> None:
         self.commands.append(("quadraticTo", float(x1), float(y1), float(x), float(y)))
         self.binary_commands.extend(struct.pack("4sffff", b"quad", float(x1), float(y1), float(x), float(y)))
 
-    def draw_image(self, img, x, y, width, height):
+    def draw_image(self, img: RGBA32Type, x: float, y: float, width: float, height: float) -> None:
         # img should be rgba pack, uint32
         assert img.dtype == numpy.uint32
         with DrawingContext.__image_id_lock:
@@ -566,7 +595,7 @@ class DrawingContext:
         self.images[str(image_id)] = img
         self.binary_commands.extend(struct.pack("4siiiffff", b"imag", img.shape[1], img.shape[0], int(image_id), float(x), float(y), float(width), float(height)))
 
-    def draw_data(self, img, x, y, width, height, low, high, color_map_data):
+    def draw_data(self, img: GrayscaleF32Type, x: float, y: float, width: float, height: float, low: float, high: float, color_map_data: typing.Optional[RGBA32Type]) -> None:
         # img should be float
         assert img.dtype == numpy.float32
         with DrawingContext.__image_id_lock:
@@ -581,47 +610,50 @@ class DrawingContext:
         if color_map_data is not None:
             self.images[str(color_map_image_id)] = color_map_data
         self.commands.append(
-            ("data", img.shape[1], img.shape[0], img, int(image_id), float(x), float(y), float(width), float(height), float(low), float(high), color_map_data, int(color_map_image_id)))
-        self.binary_commands.extend(struct.pack("4siiiffffffi", b"data", img.shape[1], img.shape[0], int(image_id), float(x), float(y), float(width), float(height), float(low), float(high), int(color_map_image_id)))
+            ("data", img.shape[1], img.shape[0], img, int(image_id), float(x), float(y), float(width), float(height),
+             float(low), float(high), color_map_data, int(color_map_image_id)))
+        self.binary_commands.extend(
+            struct.pack("4siiiffffffi", b"data", img.shape[1], img.shape[0], int(image_id), float(x), float(y),
+                        float(width), float(height), float(low), float(high), int(color_map_image_id)))
 
-    def stroke(self):
+    def stroke(self) -> None:
         self.commands.append(("stroke", ))
         self.binary_commands.extend(b"strk")
 
-    def sleep(self, duration):
+    def sleep(self, duration: float) -> None:
         self.commands.append(("sleep", float(duration)))
         self.binary_commands.extend(struct.pack("4sf", b"slep", float(duration)))
 
-    def mark_latency(self):
+    def mark_latency(self) -> None:
         self.commands.append(("latency", time.perf_counter()))
         self.binary_commands.extend(struct.pack("<4sd", b"latn", time.perf_counter()))
 
-    def message(self, text):
+    def message(self, text: str) -> None:
         self.commands.append(("message", text))
         text_encoded = text.encode("utf-8")
         self.binary_commands.extend(struct.pack("4si{}s0i".format(len(text_encoded)), b"mesg", len(text_encoded), text_encoded))
 
-    def timestamp(self, timestamp):
+    def timestamp(self, timestamp: str) -> None:
         self.commands.append(("timestamp", timestamp))
         timestamp_encoded = timestamp.encode("utf-8")
         self.binary_commands.extend(struct.pack("4si{}s0i".format(len(timestamp_encoded)), b"time", len(timestamp_encoded), timestamp_encoded))
 
-    def fill(self):
+    def fill(self) -> None:
         self.commands.append(("fill", ))
         self.binary_commands.extend(b"fill")
 
-    def fill_text(self, text, x, y, max_width=None):
+    def fill_text(self, text: str, x: float, y: float, max_width: typing.Optional[int] = None) -> None:
         text = str(text) if text is not None else str()
         self.commands.append(("fillText", text, float(x), float(y), float(max_width) if max_width else 0))
         text_encoded = text.encode("utf-8")
         self.binary_commands.extend(struct.pack("4si{}sfff".format(len(text_encoded)), b"text", len(text_encoded), text_encoded, float(x), float(y), float(max_width) if max_width else 0))
 
     @property
-    def fill_style(self):
+    def fill_style(self) -> typing.Optional[typing.Union[str, LinearGradient]]:
         raise NotImplementedError()
 
     @fill_style.setter
-    def fill_style(self, a):
+    def fill_style(self, a: typing.Optional[typing.Union[str, LinearGradient]]) -> None:
         a = a or "rgba(0, 0, 0, 0.0)"
         if isinstance(a, DrawingContext.LinearGradient):
             self.commands.extend(a.commands)
@@ -634,127 +666,165 @@ class DrawingContext:
             self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"flst", len(a_encoded), a_encoded))
 
     @property
-    def font(self):
+    def font(self) -> typing.Optional[str]:
         raise NotImplementedError()
 
     @font.setter
-    def font(self, a):
+    def font(self, a: typing.Optional[str]) -> None:
         """
             Set the text font.
 
             Supports 'normal', 'bold', 'italic', size specific as '14px', and font-family.
         """
+        assert a is not None
         self.commands.append(("font", str(a)))
         a_encoded = a.encode("utf-8")
         self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"font", len(a_encoded), a_encoded))
 
-    def __get_text_align(self):
+    @property
+    def text_align(self) -> typing.Optional[str]:
         raise NotImplementedError()
 
-    def __set_text_align(self, a):
-        """
-            Set text alignment.
+    @text_align.setter
+    def text_align(self, a: typing.Optional[str]) -> None:
+        """Set text alignment.
 
-            Valid values are 'start', 'end', 'left', 'center', 'right'. Default is 'start'.
+        Valid values are 'start', 'end', 'left', 'center', 'right'. Default is 'start'.
 
-            Default is 'start'.
+        Default is 'start'.
         """
+        assert a is not None
         self.commands.append(("textAlign", str(a)))
         a_encoded = a.encode("utf-8")
         self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"algn", len(a_encoded), a_encoded))
 
-    text_align = property(__get_text_align, __set_text_align)
-
-    def __get_text_baseline(self):
+    @property
+    def text_baseline(self) -> typing.Optional[str]:
         raise NotImplementedError()
 
-    def __set_text_baseline(self, a):
-        """
-            Set the text baseline.
+    @text_baseline.setter
+    def text_baseline(self, a: typing.Optional[str]) -> None:
+        """Set the text baseline.
 
-            Valid values are 'top', 'hanging', 'middle', 'alphabetic', 'ideographic', and 'bottom'.
+        Valid values are 'top', 'hanging', 'middle', 'alphabetic', 'ideographic', and 'bottom'.
 
-            Default is 'alphabetic'.
+        Default is 'alphabetic'.
         """
+        assert a is not None
         self.commands.append(("textBaseline", str(a)))
         a_encoded = a.encode("utf-8")
         self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"tbas", len(a_encoded), a_encoded))
 
-    text_baseline = property(__get_text_baseline, __set_text_baseline)
-
-    def __get_stroke_style(self):
+    @property
+    def stroke_style(self) -> typing.Optional[str]:
         raise NotImplementedError()
 
-    def __set_stroke_style(self, a):
+    @stroke_style.setter
+    def stroke_style(self, a: typing.Optional[str]) -> None:
         a = a or "rgba(0, 0, 0, 0.0)"
         self.commands.append(("strokeStyle", str(a)))
         a_encoded = a.encode("utf-8")
         self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"stst", len(a_encoded), a_encoded))
 
-    stroke_style = property(__get_stroke_style, __set_stroke_style)
-
-    def __get_line_width(self):
+    @property
+    def line_width(self) -> float:
         raise NotImplementedError()
 
-    def __set_line_width(self, a):
+    @line_width.setter
+    def line_width(self, a: float) -> None:
         self.commands.append(("lineWidth", float(a)))
         self.binary_commands.extend(struct.pack("4sf", b"linw", float(a)))
 
-    line_width = property(__get_line_width, __set_line_width)
-
-    def __get_line_dash(self):
+    @property
+    def line_dash(self) -> typing.Optional[int]:
         raise NotImplementedError()
 
-    def __set_line_dash(self, a):
-        """ Set the line dash. Takes a single value with the length of the dash. """
+    @line_dash.setter
+    def line_dash(self, a: typing.Optional[int]) -> None:
+        """Set the line dash. Takes a single value with the length of the dash."""
+        assert a is not None
         self.commands.append(("lineDash", float(a)))
         self.binary_commands.extend(struct.pack("4sf", b"ldsh", float(a)))
 
-    line_dash = property(__get_line_dash, __set_line_dash)
-
-    def __get_line_cap(self):
+    @property
+    def line_cap(self) -> typing.Optional[str]:
         raise NotImplementedError()
 
-    def __set_line_cap(self, a):
-        """ Set the line join. Valid values are 'square', 'round', 'butt'. Default is 'square'. """
+    @line_cap.setter
+    def line_cap(self, a: typing.Optional[str]) -> None:
+        """Set the line join. Valid values are 'square', 'round', 'butt'. Default is 'square'."""
+        assert a is not None
         self.commands.append(("lineCap", str(a)))
         a_encoded = a.encode("utf-8")
         self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"lcap", len(a_encoded), a_encoded))
 
-    line_cap = property(__get_line_cap, __set_line_cap)
-
-    def __get_line_join(self):
+    @property
+    def line_join(self) -> typing.Optional[str]:
         raise NotImplementedError()
 
-    def __set_line_join(self, a):
-        """ Set the line join. Valid values are 'round', 'miter', 'bevel'. Default is 'bevel'. """
+    @line_join.setter
+    def line_join(self, a: typing.Optional[str]) -> None:
+        """Set the line join. Valid values are 'round', 'miter', 'bevel'. Default is 'bevel'."""
+        assert a is not None
         self.commands.append(("lineJoin", str(a)))
         a_encoded = a.encode("utf-8")
         self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"lnjn", len(a_encoded), a_encoded))
 
-    line_join = property(__get_line_join, __set_line_join)
-
     class LinearGradient:
         next = 1
 
-        def __init__(self, width, height, x1, y1, x2, y2):  # pylint: disable=invalid-name
-            self.commands = []
-            self.binary_commands = []
+        def __init__(self, width: float, height: float, x1: float, y1: float, x2: float, y2: float) -> None:
+            self.commands: typing.List[typing.Sequence[typing.Any]] = []
+            self.binary_commands = bytearray()
             self.command_var = DrawingContext.LinearGradient.next
             self.commands.append(("gradient", self.command_var, float(width), float(height), float(x1), float(y1), float(x2), float(y2)))
             self.binary_commands.extend(struct.pack("4siffffff", b"grad", self.command_var, float(width), float(height), float(x1), float(y1), float(x2), float(y2)))
             DrawingContext.LinearGradient.next += 1
 
-        def add_color_stop(self, x, color):
+        def add_color_stop(self, x: float, color: str) -> None:
             self.commands.append(("colorStop", self.command_var, float(x), str(color)))
             color_encoded = color.encode("utf-8")
             self.binary_commands.extend(struct.pack("4sifi{}s0i".format(len(color_encoded)), b"grcs", self.command_var, float(x), len(color_encoded), color_encoded))
 
-    def create_linear_gradient(self, width, height, x1, y1, x2, y2):  # pylint: disable=invalid-name
+    def create_linear_gradient(self, width: float, height: float, x1: float, y1: float, x2: float, y2: float) -> DrawingContext.LinearGradient:
         gradient = DrawingContext.LinearGradient(width, height, x1, y1, x2, y2)
         return gradient
 
-    def statistics(self, stat_id):
+    def statistics(self, stat_id: str) -> None:
         self.commands.append(("statistics", str(stat_id)))
         stat_id_encoded = stat_id.encode("utf-8")
         self.binary_commands.extend(struct.pack("4si{}s0i".format(len(stat_id_encoded)), b"stat", len(stat_id_encoded), stat_id_encoded))
+
+
+class Path:
+    def __init__(self) -> None:
+        self.__path = DrawingContext()
+
+    def __deepcopy__(self, memo: typing.Dict[typing.Any, typing.Any]) -> Path:
+        path = Path()
+        path.__path = copy.deepcopy(self.__path)
+        return path
+
+    def add_commands(self, drawing_context: DrawingContext) -> None:
+        drawing_context.add(self.__path)
+
+    def move_to(self, x: float, y: float) -> None:
+        self.__path.move_to(x, y)
+
+    def line_to(self, x: float, y: float) -> None:
+        self.__path.line_to(x, y)
+
+    def close_path(self) -> None:
+        self.__path.close_path()
+
+
+def color_without_alpha(color: typing.Optional[str]) -> typing.Optional[str]:
+    return Color.Color(color).to_color_without_alpha().color_str
+
+
+def named_color_without_alpha(color: str) -> typing.Optional[str]:
+    return Color.Color(color).to_named_color_without_alpha().color_str
+
+
+def hex_color(color: typing.Union[str, Color.Color]) -> typing.Optional[str]:
+    return Color.Color(color).hex_color_str
