@@ -9,6 +9,7 @@ import enum
 import functools
 import gettext
 import logging
+import pathlib
 import typing
 import weakref
 
@@ -16,10 +17,10 @@ import weakref
 from nion.utils import Event
 from nion.utils import Geometry
 from nion.utils import Process
+from nion.ui import DrawingContext
 from nion.ui import UserInterface
 
 if typing.TYPE_CHECKING:
-    import numpy
     from nion.ui import Application
 
 
@@ -28,19 +29,34 @@ _ = gettext.gettext
 
 class ActionContext:
     def __init__(self, application: Application.BaseApplication, window: typing.Optional[Window], focus_widget: typing.Optional[UserInterface.Widget]):
-        self.application = application
-        self.window = window
+        self._application = application
+        self._window = window
         self.focus_widget = focus_widget
         self.event_type_str: typing.Optional[str] = None
+        self.parameters: typing.Dict[str, typing.Any] = dict()
+
+    @property
+    def application(self) -> Application.BaseApplication:
+        return self._application
+
+    @property
+    def window(self) -> typing.Optional[Window]:
+        return self._window
 
 
-class ActionResult(enum.IntEnum):
+class ActionStatus(enum.IntEnum):
     FINISHED = 0
     MODAL = 1
     CANCELLED = 2
     PASS = 3
 
 
+class ActionResult:
+    def __init__(self, status: ActionStatus):
+        self.status = status
+        self.results: typing.Dict[str, typing.Any] = dict()
+
+
 class ReportType(enum.IntEnum):
     NOTSET = 0
     DEBUG = 10
@@ -60,14 +76,29 @@ class Report:
         return Report(ReportType(record.levelno), record.getMessage())
 
 
+class ActionProperty:
+    pass
+
+
+class ActionStringProperty(ActionProperty):
+    def __init__(self, name: str):
+        self.name = name
+
+
+class ActionIntegerProperty(ActionProperty):
+    def __init__(self, name: str):
+        self.name = name
+
+
 class Action:
     action_id: str = str()
     action_name: str = str()
     action_role: typing.Optional[str] = None
     action_summary: typing.Optional[str] = None
     action_description: typing.Optional[str] = None
+    action_parameters: typing.List[ActionProperty] = list()
 
-    def __init__(self):
+    def __init__(self) -> None:
         self.__reports: typing.List[Report] = list()
 
     @property
@@ -83,7 +114,7 @@ class Action:
 
     def execute(self, context: ActionContext) -> ActionResult:
         """Execute the action with the context. No user interaction allowed. May be called from scripts."""
-        ...
+        raise NotImplementedError()
 
     def invoke(self, context: ActionContext) -> ActionResult:
         """Called to execute the action with the context. User interaction allowed. Typically calls `execute`."""
@@ -104,6 +135,18 @@ class Action:
     def log_record(self, record: logging.LogRecord) -> None:
         self.__reports.append(Report.from_log_record(record))
 
+    def get_string_property(self, context: ActionContext, name: str) -> str:
+        return str(context.parameters.get(name, str()))
+
+    def set_string_property(self, context: ActionContext, name: str, value: str) -> None:
+        context.parameters[name] = value
+
+    def get_int_property(self, context: ActionContext, name: str) -> int:
+        return int(context.parameters.get(name, 0))
+
+    def set_int_property(self, context: ActionContext, name: str, value: int) -> None:
+        context.parameters[name] = value
+
 
 actions: typing.Mapping[str, Action] = dict()
 
@@ -113,24 +156,39 @@ def register_action(action: Action) -> None:
     typing.cast(typing.MutableMapping[str, Action], actions)[action.action_id] = action
 
 
-action_shortcuts: typing.Mapping[str, typing.Mapping] = dict()
+_ActionShortcutsType = typing.Mapping[str, typing.Mapping[str, str]]
+
+action_shortcuts: typing.Mapping[str, typing.Mapping[str, str]] = dict()
 
 
 def add_action_shortcut(action_id: str, action_context: str, key_sequence: str) -> None:
-    typing.cast(typing.MutableMapping[str, typing.MutableMapping], action_shortcuts).setdefault(action_id, dict())[action_context] = key_sequence
+    typing.cast(typing.MutableMapping[str, typing.MutableMapping[str, str]], action_shortcuts).setdefault(action_id, dict())[action_context] = key_sequence
 
 
-def register_action_shortcuts(action_shortcuts: typing.Mapping) -> None:
+def register_action_shortcuts(action_shortcuts: _ActionShortcutsType) -> None:
     for action_id, action_shortcut_d in action_shortcuts.items():
         for action_context, key_sequence in action_shortcut_d.items():
             add_action_shortcut(action_id, action_context, key_sequence)
 
 
-def get_action_id_for_key(context: str, key) -> typing.Optional[str]:
-    for action_id, action_shortcut_d in action_shortcuts.items():
-        for action_context, key_sequence in action_shortcut_d.items():
-            if action_context == context and key.text == key_sequence:  # TODO: match the actual key sequence
-                return action_id
+def get_action_id_for_key(context: str, key: UserInterface.Key) -> typing.Optional[str]:
+    # deprecated as of 0.5.2. use get_action_for_key instead.
+    action = get_action_for_key([context], key)
+    return action.action_id if action else None
+
+
+def get_action_for_key(action_contexts: typing.Sequence[str], key: UserInterface.Key) -> typing.Optional[Action]:
+    for action_context in reversed(action_contexts):
+        for action_id, action_shortcut_d in action_shortcuts.items():
+            for key_action_context, key_sequence_str in action_shortcut_d.items():
+                if action_context == key_action_context:
+                    if isinstance(key_sequence_str, list):
+                        key_sequences = [UserInterface.KeySequence(kss) for kss in key_sequence_str]
+                    else:
+                        key_sequences = [UserInterface.KeySequence(key_sequence_str)]
+                    for key_sequence in key_sequences:
+                        if key_sequence.matches(key) == UserInterface.KeySequenceMatch.EXACT:
+                            return actions[action_id]
     return None
 
 
@@ -138,7 +196,7 @@ class Window:
     count = 0  # useful for detecting leaks in tests
 
     def __init__(self, ui: UserInterface.UserInterface, app: typing.Optional[Application.BaseApplication] = None,
-                 parent_window: Window = None, window_style: typing.Optional[str] = None,
+                 parent_window: typing.Optional[Window] = None, window_style: typing.Optional[str] = None,
                  persistent_id: typing.Optional[str] = None):
         Window.count += 1
         self.ui = ui
@@ -153,7 +211,8 @@ class Window:
         self.__shown = False
         self.__request_close = False
 
-        self.__dialogs: typing.List[weakref.ReferenceType] = list()
+        # Python 3.9+: should be weakref.ReferenceType[Dialog]
+        self.__dialogs: typing.List[typing.Any] = list()
 
         self._window_close_event = Event.Event()
 
@@ -191,12 +250,8 @@ class Window:
         self._zoom_action = None
         self._bring_to_front_action = None
 
-        # configure the event loop object
-        logger = logging.getLogger()
-        old_level = logger.level
-        logger.setLevel(logging.INFO)
-        self.__event_loop = asyncio.new_event_loop()  # outputs a debugger message!
-        logger.setLevel(old_level)
+        # configure the event loop object. for backwards compatibility only.
+        self.__event_loop = asyncio.get_event_loop_policy().get_event_loop()
 
         if app:
             app._window_created(self)
@@ -208,7 +263,6 @@ class Window:
         self._window_close_event.fire(self)
         self._window_close_event = typing.cast(Event.Event, None)
         self.on_close = None
-        Process.close_event_loop(self.__event_loop)
         self.__event_loop = typing.cast(asyncio.AbstractEventLoop, None)
         self.ui.destroy_document_window(self.__document_window)  # close the ui window
         self.__document_window = typing.cast(UserInterface.Window, None)
@@ -276,7 +330,7 @@ class Window:
              },
         ]
 
-        self.build_menu(None, typing.cast(typing.List[typing.Dict], menu_descriptions))
+        self.build_menu(None, typing.cast(typing.List[typing.Dict[str, typing.Any]], menu_descriptions))
 
     def _adjust_menus(self) -> None:
         # called when key may be shortcut. does not work for sub-menus.
@@ -287,6 +341,14 @@ class Window:
         if self.app:
             self.app.exit()
 
+    @property
+    def window_file_path(self) -> typing.Optional[pathlib.Path]:
+        return self.__document_window.window_file_path
+
+    @window_file_path.setter
+    def window_file_path(self, value: typing.Optional[pathlib.Path]) -> None:
+        self.__document_window.window_file_path = value
+
     def request_close(self) -> None:
         self.__document_window.request_close()
 
@@ -316,7 +378,7 @@ class Window:
         if self.__request_close:
             self.request_close()
 
-    def exec_action_events(self, event_type_str: str, **kwargs) -> bool:
+    def exec_action_events(self, event_type_str: str, **kwargs: typing.Any) -> bool:
         action_context = None
         for action in self.__modal_actions:
             if not action_context:
@@ -326,10 +388,10 @@ class Window:
                     setattr(action_context, k, v)
             action_result = action.event(action_context)
             # finished or cancelled will remove the action
-            if action_result not in {ActionResult.MODAL, ActionResult.PASS}:
+            if action_result.status not in {ActionStatus.MODAL, ActionStatus.PASS}:
                 self.__modal_actions.remove(action)
             # modal, finished, or cancelled will stop iterating and return True
-            if action_result != ActionResult.PASS:
+            if action_result.status != ActionStatus.PASS:
                 for report in action.reports:
                     self.display_report(report)
                 return True
@@ -337,7 +399,10 @@ class Window:
         return False
 
     def _close_dialogs(self) -> None:
-        for weak_dialog in self.__dialogs:
+        # each window may have its own dialogs, but for the most part, this is used for the top level
+        # windows. request close will trigger close which will remove items from __dialogs, so be sure
+        # to copy the list so that it doesn't get modified while iterating.
+        for weak_dialog in list(self.__dialogs):
             dialog = typing.cast("Window", weak_dialog())
             if dialog:
                 try:
@@ -355,7 +420,7 @@ class Window:
     def register_dialog(self, dialog: Window) -> None:
         old_on_close = dialog.on_close
 
-        def close_dialog():
+        def close_dialog() -> None:
             self.__dialogs.remove(weakref.ref(dialog))
             if callable(old_on_close):
                 old_on_close()
@@ -416,8 +481,10 @@ class Window:
             self._adjust_menus()
         return False
 
-    def drag(self, mime_data: UserInterface.MimeData, thumbnail: numpy.ndarray, hot_spot_x: int, hot_spot_y: int) -> None:
-        self.__document_window.root_widget.drag(mime_data, thumbnail, hot_spot_x, hot_spot_y)
+    def drag(self, mime_data: UserInterface.MimeData, thumbnail: typing.Optional[DrawingContext.RGBA32Type], hot_spot_x: int, hot_spot_y: int) -> None:
+        root_widget = self.__document_window.root_widget
+        if root_widget:
+            root_widget.drag(mime_data, thumbnail, hot_spot_x, hot_spot_y)
 
     @property
     def title(self) -> str:
@@ -427,13 +494,16 @@ class Window:
     def title(self, value: str) -> None:
         self.__document_window.title = value
 
-    def get_file_paths_dialog(self, title: str, directory: str, filter: str, selected_filter: str = None) -> typing.Tuple[typing.List[str], str, str]:
+    def get_file_paths_dialog(self, title: str, directory: str, filter: str,
+                              selected_filter: typing.Optional[str] = None) -> typing.Tuple[typing.List[str], str, str]:
         return self.__document_window.get_file_paths_dialog(title, directory, filter, selected_filter)
 
-    def get_file_path_dialog(self, title: str, directory: str, filter: str, selected_filter: str = None) -> typing.Tuple[typing.List[str], str, str]:
+    def get_file_path_dialog(self, title: str, directory: str, filter: str,
+                             selected_filter: typing.Optional[str] = None) -> typing.Tuple[typing.List[str], str, str]:
         return self.__document_window.get_file_path_dialog(title, directory, filter, selected_filter)
 
-    def get_save_file_path(self, title: str, directory: str, filter: str, selected_filter: str = None) -> typing.Tuple[str, str, str]:
+    def get_save_file_path(self, title: str, directory: str, filter: str,
+                           selected_filter: typing.Optional[str] = None) -> typing.Tuple[str, str, str]:
         return self.__document_window.get_save_file_path(title, directory, filter, selected_filter)
 
     def create_dock_widget(self, widget: UserInterface.Widget, panel_id: str, title: str,
@@ -477,16 +547,19 @@ class Window:
     def dock_widgets(self) -> typing.Sequence[UserInterface.DockWidget]:
         return self.__document_window.dock_widgets
 
-    def show(self, *, size: Geometry.IntSize = None, position: Geometry.IntPoint = None) -> None:
+    def show(self, *, size: typing.Optional[Geometry.IntSize] = None, position: typing.Optional[Geometry.IntPoint] = None) -> None:
         self.__document_window.show(size=size, position=position)
 
-    def add_menu(self, title: str, menu_id: str = None) -> UserInterface.Menu:
+    def activate(self) -> None:
+        self.__document_window.activate()
+
+    def add_menu(self, title: str, menu_id: typing.Optional[str] = None) -> UserInterface.Menu:
         return self.__document_window.add_menu(title, menu_id)
 
-    def insert_menu(self, title: str, before_menu: UserInterface.Menu, menu_id: str = None) -> UserInterface.Menu:
+    def insert_menu(self, title: str, before_menu: UserInterface.Menu, menu_id: typing.Optional[str] = None) -> UserInterface.Menu:
         return self.__document_window.insert_menu(title, before_menu, menu_id)
 
-    def create_sub_menu(self, title: str = None, menu_id: str = None) -> UserInterface.Menu:
+    def create_sub_menu(self, title: typing.Optional[str] = None, menu_id: typing.Optional[str] = None) -> UserInterface.Menu:
         return self.ui.create_sub_menu(self.__document_window, title, menu_id)
 
     def create_context_menu(self) -> UserInterface.Menu:
@@ -521,7 +594,7 @@ class Window:
         assert self.app
         self.app.exit()
 
-    def _dispatch_any_to_focus_widget(self, method: str, *args, **kwargs) -> bool:
+    def _dispatch_any_to_focus_widget(self, method: str, *args: typing.Any, **kwargs: typing.Any) -> bool:
         focus_widget = self.focus_widget
         if focus_widget and focus_widget._dispatch_any(method, *args, **kwargs):
             return True
@@ -537,7 +610,7 @@ class Window:
             return True
         return False
 
-    def build_menu(self, menu: typing.Optional[UserInterface.Menu], menu_descriptions: typing.Sequence[typing.Mapping]) -> None:
+    def build_menu(self, menu: typing.Optional[UserInterface.Menu], menu_descriptions: typing.Sequence[typing.Mapping[str, typing.Any]]) -> None:
         for item_d in menu_descriptions:
             item_type = item_d["type"]
             if item_type == "menu":
@@ -591,16 +664,28 @@ class Window:
                 checked = action and action.is_checked(action_context)
                 menu_action.apply_state(UserInterface.MenuItemState(title=title, enabled=enabled, checked=checked))
 
+    def is_action_enabled(self, action_id: str, action_context: ActionContext) -> bool:
+        action = actions.get(action_id)
+        if action and action.is_enabled(action_context):
+            return True
+        return False
+
+    def add_action_to_menu_if_enabled(self, menu: UserInterface.Menu, action_id: str, action_context: ActionContext) -> typing.Optional[Action]:
+        action = actions.get(action_id)
+        if action and action.is_enabled(action_context):
+            self.add_action_to_menu(menu, action_id, action_context)
+        return action
+
     def add_action_to_menu(self, menu: UserInterface.Menu, action_id: str, action_context: ActionContext) -> typing.Optional[Action]:
         action = actions.get(action_id)
         if action:
             key_sequence = action_shortcuts.get(action_id, dict()).get("window")
             assert menu is not None
 
-            def perform_action():
+            def perform_action() -> None:
                 self.perform_action_in_context(action_id, action_context)
 
-            def queue_perform_action():
+            def queue_perform_action() -> None:
                 # delay execution to ensure menu closes properly
                 # this would manifest itself by export dialog crashes in nionswift.
                 self.queue_task(perform_action)
@@ -614,14 +699,26 @@ class Window:
             menu_action.apply_state(UserInterface.MenuItemState(title=title, enabled=enabled, checked=checked))
         return action
 
-    def perform_action(self, action_id: str) -> None:
-        self.perform_action_in_context(action_id, self._get_action_context())
+    def execute_action(self, action_id: str, action_context: typing.Optional[ActionContext] = None,
+                       parameters: typing.Optional[typing.Dict[str, typing.Any]] = None) -> ActionResult:
+        context = action_context or self._get_action_context()
+        if parameters:
+            context.parameters.update(parameters)
+        action = actions[action_id]
+        return action.execute(context)
 
-    def perform_action_in_context(self, action_id: str, action_context: ActionContext) -> None:
-        action = actions.get(action_id)
+    def perform_action(self, action: typing.Union[Action, str]) -> None:
+        self.perform_action_in_context(action, self._get_action_context())
+
+    def perform_action_in_context(self, action_or_action_id: typing.Union[Action, str], action_context: ActionContext) -> None:
+        action: typing.Optional[Action]
+        if isinstance(action_or_action_id, Action):
+            action = action_or_action_id
+        else:
+            action = actions.get(str(action_or_action_id))
         if action and action not in self.__modal_actions:
             action.clear()
-            if action.invoke(action_context) == ActionResult.MODAL:
+            if action.invoke(action_context).status == ActionStatus.MODAL:
                 self.__modal_actions.append(action)
             for report in action.reports:
                 self.display_report(report)
@@ -651,7 +748,7 @@ class Window:
         if hasattr(self, menu_item_state_method):
             menu_item_state = getattr(self, menu_item_state_method)()
             if menu_item_state:
-                return menu_item_state
+                return typing.cast(UserInterface.MenuItemState, menu_item_state)
         if hasattr(self, handle_method):
             return UserInterface.MenuItemState(title=None, enabled=True, checked=False)
         return None
@@ -782,22 +879,25 @@ class AboutBoxAction(Action):
     action_name = _("About...")
     action_role = "about"
 
+    def execute(self, context: ActionContext) -> ActionResult:
+        raise NotImplementedError()
+
     def invoke(self, context: ActionContext) -> ActionResult:
         show_about_box = getattr(context.window, "show_about_box", None)
         show_about_box = show_about_box or getattr(context.application, "show_about_box", None)
         if callable(show_about_box):
             show_about_box()
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
 
 class BringToFrontAction(Action):
     action_id = "window.bring_to_front"
     action_name = _("Bring to Front")
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("bring_to_front")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         return hasattr(context.window, "bring_to_front")
@@ -807,20 +907,20 @@ class CloseWindowAction(Action):
     action_id = "window.close"
     action_name = _("Close Window")
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window.request_close()
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
 
 class CopyAction(Action):
     action_id = "window.copy"
     action_name = _("Copy")
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("handle_copy")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         return hasattr(context.window, "handle_copy")
@@ -830,10 +930,10 @@ class CutAction(Action):
     action_id = "window.cut"
     action_name = _("Cut")
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("handle_cut")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         return hasattr(context.window, "handle_cut")
@@ -843,10 +943,10 @@ class DeleteAction(Action):
     action_id = "window.delete"
     action_name = _("Delete")
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("handle_delete")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         return hasattr(context.window, "handle_delete")
@@ -857,19 +957,19 @@ class ExitAction(Action):
     action_name = _("Exit")
     action_role = "quit"
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         context.application.exit()
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
 
 class MinimizeAction(Action):
     action_id = "window.minimize"
     action_name = _("Minimize")
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("handle_minimize")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         return hasattr(context.window, "handle_minimize")
@@ -879,10 +979,13 @@ class PageSetupAction(Action):
     action_id = "window.page_setup"
     action_name = _("Page Setup")
 
+    def execute(self, context: ActionContext) -> ActionResult:
+        raise NotImplementedError()
+
     def invoke(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("handle_page_setup")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         return hasattr(context.window, "handle_page_setup")
@@ -892,10 +995,10 @@ class PasteAction(Action):
     action_id = "window.paste"
     action_name = _("Paste")
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("handle_paste")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         return hasattr(context.window, "handle_paste")
@@ -906,22 +1009,28 @@ class PreferencesAction(Action):
     action_name = _("Preferences...")
     action_role = "preferences"
 
+    def execute(self, context: ActionContext) -> ActionResult:
+        raise NotImplementedError()
+
     def invoke(self, context: ActionContext) -> ActionResult:
         open_preferences = getattr(context.window, "open_preferences", None)
         open_preferences = open_preferences or getattr(context.application, "open_preferences", None)
         if callable(open_preferences):
             open_preferences()
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
 
 class PrintAction(Action):
     action_id = "window.print"
     action_name = _("Print...")
 
+    def execute(self, context: ActionContext) -> ActionResult:
+        raise NotImplementedError()
+
     def invoke(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("handle_print")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         return hasattr(context.window, "handle_print")
@@ -931,10 +1040,10 @@ class RedoAction(Action):
     action_id = "window.redo"
     action_name = _("Redo")
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("handle_redo")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         return hasattr(context.window, "handle_redo")
@@ -944,10 +1053,10 @@ class SelectAllAction(Action):
     action_id = "window.select_all"
     action_name = _("Select All")
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("handle_select_all")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         if context.window:
@@ -959,10 +1068,10 @@ class UndoAction(Action):
     action_id = "window.undo"
     action_name = _("Undo")
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("handle_undo")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         return hasattr(context.window, "handle_undo")
@@ -972,10 +1081,10 @@ class ZoomAction(Action):
     action_id = "window.zoom"
     action_name = _("Zoom")
 
-    def invoke(self, context: ActionContext) -> ActionResult:
+    def execute(self, context: ActionContext) -> ActionResult:
         if context.window:
             context.window._dispatch_any_to_focus_widget("handle_zoom")
-        return ActionResult.FINISHED
+        return ActionResult(ActionStatus.FINISHED)
 
     def is_enabled(self, context: ActionContext) -> bool:
         return hasattr(context.window, "handle_zoom")
